#define WIDTH  320
#define HEIGHT 240
#define DX 4 / WIDTH / 3
#define DY 1 / HEIGHT

// Complex operations
inline float2 cmul(float2 a, float2 b){return (float2)(a.x*b.x - a.y*b.y, a.x*b.y + a.y*b.x);}
inline float2 cdiv(float2 a, float2 b){return (float2)((a.x*b.x + a.y*b.y)/(b.x*b.x + b.y*b.y), (a.y*b.x - a.x*b.y)/(b.x*b.x + b.y*b.y));}
inline float2 csqrt(float2 a){
    float r = sqrt(sqrt(dot(a, a)));
    float theta = 0.5f * atan2(a.y, a.x);
    return (float2)(r * cos(theta), r * sin(theta));
}
inline float2 ccbrt(float2 a)
{
	float r, theta;
	theta = atan2(a.y, a.x) / 3.f;
	r = pow(sqrt(dot(a, a)), 1.f / 3.f);
	return (float2)(r * cos(theta), (r * sin(theta)));
}

float3 reflect(float3 r, float3 n)
{
    return r - 2*(dot(r,n))*n;
}

float intersect_quad_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    __global __read_only float*  quad_pos,
    __global __read_only float*  quad_mat,
    __global __read_only float*  quad_prp,
    __global __read_only int* quad_tex,
    int quads,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < quads; i++){
        float3  x = (float3)(quad_pos[i * 14 + 0], quad_pos[i * 14 + 1], quad_pos[i * 14 + 2]);
        float3  n = (float3)(quad_pos[i * 14 + 3], quad_pos[i * 14 + 4], quad_pos[i * 14 + 5]);
        float3 e1 = (float3)(quad_pos[i * 14 + 6], quad_pos[i * 14 + 7], quad_pos[i * 14 + 8]);
        float3 e2 = (float3)(quad_pos[i * 14 + 9], quad_pos[i * 14 +10], quad_pos[i * 14 +11]);
        float w = quad_pos[i * 14 + 12];
        float h = quad_pos[i * 14 + 13];

        float tmp = dot(n, rd);
        float t = (tmp != 0.f) ? dot(n, (x - ro)) / tmp : 9000.f;
        if(t < dist && t > 0.f){
            float3 e = ro + t * rd - x;
            float p1 = dot(e, e1) / w;
            float p2 = dot(e, e2) / h;
            if((0.f <= p1 && p1 <= w) && (0.f <= p2 && p2 <= h)){
                dist = t;
                *nml = n;
                *prp = (float3)(quad_prp[i * 3], quad_prp[i * 3 + 1], quad_prp[i * 3 + 2]);
                if(quad_tex[3 * i] == -1)
                    *col = (float3)(quad_mat[i * 3], quad_mat[i * 3 + 1], quad_mat[i * 3 + 2]);
                else{

                    size_t offset = texprops[3 * quad_tex[3 * i]];
                    size_t u, v;
                    u = (size_t)(p1 * 4) % quad_tex[3 * i + 1];
                    v = (size_t)(p2 * 4) % quad_tex[3 * i + 2];
                    offset += 3 * (v * quad_tex[3 * i + 1] + u);
                    *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
                }
            }
        }
    }
    return dist;
}

float intersect_quad(
    __global __read_only float*  quad_pos,
    int quads,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < quads; i++){
        float3 x = (float3)(quad_pos[i * 14 + 0], quad_pos[i * 14 + 1], quad_pos[i * 14 + 2]);
        float3 n = (float3)(quad_pos[i * 14 + 3], quad_pos[i * 14 + 4], quad_pos[i * 14 + 5]);
        float3 e1 = (float3)(quad_pos[i * 14 + 6], quad_pos[i * 14 + 7], quad_pos[i * 14 + 8]);
        float3 e2 = (float3)(quad_pos[i * 14 + 9], quad_pos[i * 14 +10], quad_pos[i * 14 +11]);
        float w = quad_pos[i * 14 + 12];
        float h = quad_pos[i * 14 + 13];

        float tmp = dot(n, rd);
        float t = (tmp != 0.f) ? dot(n, (x - ro)) / tmp : 9000.f;
        if(t < dist && t > 0.f){
            float3 v = ro + t * rd - x;
            float p1 = dot(v, e1) / w;
            float p2 = dot(v, e2) / h;
            if((0.f <= p1 && p1 <= w) && (0.f <= p2 && p2 <= h)){
                dist = t;
            }
        }
    }
    return dist;
}

float intersect_sphere_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    __global __read_only float* sphere_pos,
    __global __read_only float* sphere_mat,
    __global __read_only float* sphere_prp,
    __global __read_only int*   sphere_tex,
    int spheres,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < 6; i++){
        float3 x = (float3)(sphere_pos[i * 4 + 0], sphere_pos[i * 4 + 1], sphere_pos[i * 4 + 2]);
        float r = sphere_pos[i * 4 + 3];
        float3 rc = ro - x;
        float c = dot(rc, rc) - (r*r);
        float b = dot(rd, rc);
        float d = b*b - c;
        float t = -b - sqrt(fabs(d));
        float s = step(0.0, min(t,d)) * step(t, dist);
        if(t < dist && min(t,d) > 0.f){
            float3 p = ro + rd * t - x;
            dist = t;
            *nml = -p / r;
            p /= r;
            //*nml = (x - (ro + rd * t)) / r;
            *prp = (float3)(sphere_prp[i * 3 + 0], sphere_prp[i * 3 + 1], sphere_prp[i * 3 + 2]);
            if(sphere_tex[3 * i] == -1)
                *col = (float3)(sphere_mat[i * 3 + 0], sphere_mat[i * 3 + 1], sphere_mat[i * 3 + 2]);
            else{
                float ru = (1.f - (atan2(p.z, p.x) + 3.14159265f) / 6.283185f);
                float rv = ((asin(p.y) + 1.570796326f) / 3.14159265f);
                ru = clamp(ru, 0.001, 0.999);
                rv = clamp(rv, 0.001, 0.999);
                size_t u = (size_t)(ru * sphere_tex[3 * i + 1]);
                size_t v = (size_t)(rv * sphere_tex[3 * i + 2]);
                size_t offset = texprops[3 * sphere_tex[3 * i]];
                offset += 3 * (v * sphere_tex[3 * i + 1] + u);
                *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
            }
        }
    }
    return dist;
}

float intersect_sphere(
    __global __read_only float* sphere_pos,
    int spheres,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < 6; i++){
        float3 x = (float3)(sphere_pos[i * 4 + 0], sphere_pos[i * 4 + 1], sphere_pos[i * 4 + 2]);
        float r = sphere_pos[i * 4 + 3];
        float3 rc = ro - x;
        float c = dot(rc, rc) - (r*r);
        float b = dot(rd, rc);
        float d = b*b - c;
        float t = -b - sqrt(fabs(d));
        float s = step(0.0, min(t,d)) * step(t, dist);
        if(t < dist && min(t,d) > 0.f){
            dist = t;
        }
    }
    return dist;
}

float intersect_curve_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    float3 curve_pos,
    float w,
    float h,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    float3 ri = ro - curve_pos;
    float q = -h / (w * w);
    float a = (rd.x * rd.x + rd.z * rd.z) * q;
    float b = (2 * q * (ri.x * rd.x + ri.z * rd.z) - rd.y);
    float c = (ri.x * ri.x + ri.z * ri.z) * q - ri.y + h;

    float d = b * b - 4.f * a * c;
    float t;
    if(d > 0.f){
        float t1 = (-b - sqrt(fabs(d))) / (2 * a);
        float t2 = (-b + sqrt(fabs(d))) / (2 * a);  // will this always have a higher value than t1?

        if(t1 > 0.f && t2 > 0.f)
            t = fmin(t1, t2);
        else if(fmax(t1, t2) > 0.f)
            t = fmax(t1, t2);
        else
            t = 10000.0;

        float3 rx = ro + rd * t - curve_pos;
        float3 e1 = (float3)(  w, 0.0, 0.0);
        float3 e2 = (float3)(0.0, 0.0,   w);
        float p1 = dot(rx, e1) / 4.0;
        float p2 = dot(rx, e2) / 4.0;

        if(t < dist && t > 0.f){
            *nml = normalize((float3)(-2.f * rx.x * q, 1.f, -2.f * rx.z * q));
            *col = (float3)(1.00, 0.50, 0.25);
            *prp = (float3)(1.00, 0.00, 0.00);
            dist = t;

            size_t offset = texprops[3];
            size_t u, v;
            u = (size_t)(p1 * 4 + 3) % 16;
            v = (size_t)(p2 * 4 + 3) % 16;
            offset += 3 * (v * 16 + u);
            *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
        }
    }
    return dist;
}

float intersect_curve(
    float3 curve_pos,
    float w,
    float h,
    float3 ro,
    float3 rd,
    float dist
){
    float3 ri = ro - curve_pos;
    float q = -h / (w * w);
    float a = (rd.x * rd.x + rd.z * rd.z) * q;
    float b = (2 * q * (ri.x * rd.x + ri.z * rd.z) - rd.y);
    float c = (ri.x * ri.x + ri.z * ri.z) * q - ri.y + h;

    float d = b * b - 4 * a * c;
    float t;
    if(d > 0.f){
        float t1 = (-b - sqrt(fabs(d))) / (2 * a);
        float t2 = (-b + sqrt(fabs(d))) / (2 * a);  // will this always have a higher value than t1?

        if(t1 > 0.f && t2 > 0.f)
            t = fmin(t1, t2);
        else if(fmax(t1, t2) > 0.f)
            t = fmax(t1, t2);
        else
            t = 10000.0;

        float3 rx = ro + rd * t - curve_pos;
        float3 e1 = (float3)(4.0, 0.0, 0.0);
        float3 e2 = (float3)(0.0, 0.0, 4.0);
        float p1 = dot(rx, e1) / 4.0;
        float p2 = dot(rx, e2) / 4.0;
        if(t < dist && t > 0.f){
            dist = t;
        }
    }
    return dist;
}

float intersect_collapse_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    float3 obj_pos,
    float w,
    float h,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    float3 ri = ro - obj_pos;
    float xr2 = (rd.x * rd.x + rd.z * rd.z) * 1.f;
    float xir = (ri.x * rd.x + ri.z * rd.z) * 2.f;
    float xi2 = (ri.x * ri.x + ri.z * ri.z) * 1.f;
    float yih = (ri.y + h / (w * w));
    float a = rd.y * xr2;
    float b = rd.y * xir + yih * xr2;
    float c = rd.y * xi2 + yih * xir;
    float d =  yih * xi2 - h;
    a = sign(a) * fmax(fabs(a), 0.001);
    if(a == 0.f)
        return dist;

    float2 d0 = (float2)(b * b - 3 * a * c, 0.f);
    float2 d1 = (float2)(2 * b * b * b - 9 * a * b * c + 27 * a * a * d, 0.f);
    float2 sr = csqrt(d1 * d1 - 4 * d0 * d0 * d0);
    float2 cr = 0.5f * (d1 - sr);
    float2 ep = (float2)(-0.5, 0.5 * sqrt(3.f));
    if((fabs(cr.x) + fabs(cr.y)) < 0.1f)
        cr = 0.5f * (d1 + sr);
    cr = ccbrt(cr);

    float t = 9999.9;
    float q = -1.f / (3.f * a);
    float2 xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.1f && xk.x > 0.f)
        t = xk.x;
    cr = cmul(ep, cr);
    xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.1f && xk.x > 0.f && xk.x < t)
        t = xk.x;
    cr = cmul(ep, cr);
    xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.1f && xk.x > 0.f && xk.x < t)
        t = xk.x;

    float3 rx = ri + rd * t;
    float3 e1 = (float3)(2*w, 0.0, 0.0);
    float3 e2 = (float3)(0.0, 0.0, 2*w);
    float p1 = dot(rx, e1) / (2 * w) + 126;
    float p2 = dot(rx, e2) / (2 * w) + 126;
    if(t < dist && rx.x * rx.x + rx.z * rx.z < w){
        float tmp = rx.x * rx.x + rx.z * rx.z;
        float y_x = 2.f * rx.x * h / (tmp * tmp + 0.0001);
        float y_z = 2.f * rx.z * h / (tmp * tmp + 0.0001);
        *nml = normalize((float3)(y_x, 1.f, y_z));
        *col = (float3)(1.00, 0.50, 0.25);
        *prp = (float3)(1.00, 0.00, 0.00);
        dist = t;

        size_t u, v;
        size_t offset = texprops[3];
        u = (size_t)(p1 * 4) % 16;
        v = (size_t)(p2 * 4) % 16;
        offset += 3 * (v * 16 + u);
        *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
    }

    return dist;
}

float intersect_collapse(
    float3 obj_pos,
    float w,
    float h,
    float3 ro,
    float3 rd,
    float dist
){
    float3 ri = ro - obj_pos;
    float xr2 = (rd.x * rd.x + rd.z * rd.z) * 1.f;
    float xir = (ri.x * rd.x + ri.z * rd.z) * 2.f;
    float xi2 = (ri.x * ri.x + ri.z * ri.z) * 1.f;
    float yih = (ri.y + h / (w * w));
    float a = rd.y * xr2;
    float b = rd.y * xir + yih * xr2;
    float c = rd.y * xi2 + yih * xir;
    float d =  yih * xi2 - h;
    if(a == 0.f)
        return dist;

    float2 d0 = (float2)(b * b - 3 * a * c, 0.f);
    float2 d1 = (float2)(2 * b * b * b - 9 * a * b * c + 27 * a * a * d, 0.f);
    float2 sr = csqrt(d1 * d1 - 4 * d0 * d0 * d0);
    float2 cr = 0.5f * (d1 - sr);
    float2 ep = (float2)(-0.5f, 0.5f * sqrt(3.f));
    if((fabs(cr.x) + fabs(cr.y)) == 0.f)
        cr = 0.5f * (d1 + sr);
    cr = ccbrt(cr);

    float t = 9999.9;
    float q = -1.f / (3.f * a);
    float2 xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.0001f && xk.x > 0.f)
        t = xk.x;
    cr = cmul(ep, cr);
    xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.0001f && xk.x > 0.f && xk.x < t)
        t = xk.x;
    cr = cmul(ep, cr);
    xk = q * ((float2)(b, 0.f) + cr + cdiv(d0, cr));
    if(fabs(xk.y) < 0.0001f && xk.x > 0.f && xk.x < t)
        t = xk.x;
    if(t > 9000.f)
        return dist;

    // check
    float3 rx = ri + rd * t;
    if(t < dist && rx.x * rx.x + rx.z * rx.z < w){
        dist = t;
    }

    return dist;
}

float intersect_sphere_special(
    float h,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    float3 x = (float3)(22.f, 1.f - 0.25 * sqrt(-h), 298.f);
    float r = 0.1 * sqrt(-h);
    float3 rc = ro - x;
    float c = dot(rc, rc) - (r*r);
    float b = dot(rd, rc);
    float d = b*b - c;
    float t = -b - sqrt(fabs(d));
    float s = step(0.0, min(t,d)) * step(t, dist);
    if(t < dist && min(t,d) > 0.f){
        float3 p = ro + rd * t - x;
        dist = t;
        *nml = -p / r;
        *prp = (float3)(0.75, 0.25, 0.00);
        *col = (float3)(0.10, 0.10, 0.10);
    }
    return dist;
}

float intersect_sphere_beer(
    float3* col,
    float3 ro,
    float3 rd,
    float dist
){
    float3 x = (float3)(22.f, 5.f, 298.f);
    float3 rc = ro - x;
    float c = dot(rc, rc) - 4.f;
    float b = dot(rd, rc);
    float d = b*b - c;
    float t0 = -b - sqrt(fabs(d));
    float t1 = -b + sqrt(fabs(d));
    if(t1 < dist && min(t0,d) > 0.f){
        float t = t1 - t0;
        float k = exp(-9.0f * t);
        //float3 absorb = (float3)(1.0, expf(-0.1 * t), expf(-0.1 * t));
        *col = (*col) * (float3)(1.0, k, k);
    }

    return dist;
}

float3 light_rays(
    __global __read_only float* sphere_pos,
    __global __read_only float* quad_pos,
    __global __read_only float* light_pos,
    __global __read_only float* light_col,
    int spheres, int quads, int lights,
    float3 ro,
    float c_w, float c_h, float c_y
){
    float3 nil = (float3)(0, 0, 0);
    float3 l_i = (float3)(0, 0, 0);
    for(int i = 0; i < lights; i++){
        float dist = 9999.f;
        float3 lo = (float3)(light_pos[i * 3 + 0], light_pos[i * 3 + 1], light_pos[i * 3 + 2]);
        float3 lc = (float3)(light_col[i * 3 + 0], light_col[i * 3 + 1], light_col[i * 3 + 2]);
        float3 ld = normalize(lo - ro);
        dist = intersect_quad(quad_pos, quads, ro, ld, dist);
        dist = intersect_sphere(sphere_pos, spheres, ro, ld, dist);
        //dist = intersect_curve((float3)(22.f, c_y, 298.f), c_w, c_h, ro, ld, dist);
        //dist = intersect_collapse((float3)(22.f, c_y, 298.f), c_w, c_h, ro, ld, dist);
        //dist = intersect_sphere_special(c_h, &nil, &nil, &nil, ro, ld, dist);

        //dist = intersect_sphere_beer(&lc, ro, ld, dist);

        float3 d = ro - lo;
        float x = sqrt(dot(d, d));
        float s = step(x, dist);
        l_i = l_i + mix((float3)(0.0f, 0.00f, 0.00f), 1.25 * lc * pow(0.99, x), s);
    }
    l_i = clamp(l_i, 0.1, 1.0);
    return l_i;
}

__kernel void ray_tracer(
    __global __write_only short* pixels,
    __global __read_only float* textures,
    __global __read_only size_t* texprops,
    __global __read_only float* sphere_pos,
    __global __read_only float* sphere_mat,
    __global __read_only float* sphere_prp,
    __global __read_only int* sphere_tex,
    __global __read_only float* quad_pos,
    __global __read_only float* quad_mat,
    __global __read_only float* quad_prp,
    __global __read_only int* quad_tex,
    __global __read_only float* light_pos,
    __global __read_only float* light_col,
    int spheres, int quads, int lights,
    float x_x, float x_y, float x_z,
    float r_x, float r_y, float r_z,
    int maxBounce,
    float c_w, float c_h, float c_y
)
{
    int x = get_global_id(0);
    int y = get_global_id(1);
    int idx = x + y * WIDTH;
    c_h = -c_h * c_h;

    r_x = 0.60f;

    float3 ro = (float3)(x_x, x_y, x_z);
    float3 rd = (float3)(-1.f + 2.f*DX*x, -1.f + 2.f*DY*y, 1.f);
    // rd = (float3)(rd.x, rd.y * cos(r_x) - rd.z * sin(r_x), rd.z * cos(r_x) + rd.y * sin(r_x));
    rd = (float3)(rd.x * cos(r_y) + rd.z * sin(r_y), rd.y, rd.z * cos(r_y) - rd.x * sin(r_y));
    rd = normalize(rd);

    float3 f_i = (float3)(1.f, 1.f, 1.f);  // inverse accumulated light
    float3 nml, col, mat = (float3)(0.f, 0.f, 0.f);
    float3 prp, l_i;
    for(int i = maxBounce - 1; i >= 0; i--){
        col = (float3)(0.f, 0.f, 0.f);
        float dist = 9999.f;
        float3 curve_pos = (float3)(22.f, c_y, 298.f);
        dist = intersect_quad_tex(  textures, texprops,   quad_pos,   quad_mat,   quad_prp,   quad_tex,   quads, &nml, &col, &prp, ro, rd, dist);
        dist = intersect_sphere_tex(textures, texprops, sphere_pos, sphere_mat, sphere_prp, sphere_tex, spheres, &nml, &col, &prp, ro, rd, dist);
        //dist = intersect_collapse_tex( textures, texprops,  curve_pos, c_w, c_h, &nml, &col, &prp, ro, rd, dist);
        //dist = intersect_sphere_special(c_h, &nml, &col, &prp, ro, rd, dist);

        ro = ro + dist * rd;
        rd = reflect(rd, nml);
        ro = ro + 0.01 * rd;

        //l_i = light_rays(sphere_pos, quad_pos, light_pos, light_col, spheres, quads, lights, ro);
        l_i = light_rays(sphere_pos, quad_pos, light_pos, light_col, spheres, quads, lights, ro, c_w, c_h, c_y);
        mat = mat + prp.x * col * f_i * l_i;
        f_i = f_i * l_i * prp.y;
    }
    unsigned char r = 31.f * clamp(mat.x, 0.0, 1.0);
    unsigned char g = 63.f * clamp(mat.y, 0.0, 1.0);
    unsigned char b = 31.f * clamp(mat.z, 0.0, 1.0);
    pixels[idx] = ( r << 11 ) | ( g << 5 ) | b;
}


__kernel void ray_tracer_ssaa(
    __global __write_only short* pixels,
    __global __read_only float* textures,
    __global __read_only size_t* texprops,
    __global __read_only float* sphere_pos,
    __global __read_only float* sphere_mat,
    __global __read_only float* sphere_prp,
    __global __read_only int* sphere_tex,
    __global __read_only float* quad_pos,
    __global __read_only float* quad_mat,
    __global __read_only float* quad_prp,
    __global __read_only int* quad_tex,
    __global __read_only float* light_pos,
    __global __read_only float* light_col,
    int spheres, int quads, int lights,
    float x_x, float x_y, float x_z,
    float r_x, float r_y, float r_z,
    int maxBounce,
    float c_w, float c_h, float c_y
)
{
    int x = get_global_id(0);
    int y = get_global_id(1);
    int idx = x + y * WIDTH;
    float3 aa_cols = (float3)(0.f, 0.f, 0.f);
    for(int ii = 0; ii < 2; ii++){
        for(int jj = 0; jj < 2; jj++){
            float3 ro = (float3)(x_x, x_y, x_z);
            float3 rd = (float3)(-1.f + .99f*DX*(2.0*x + ii), -1.f + .99f*DY*(2.0*y + jj), 1.f);

            rd = (float3)(rd.x * cos(r_y) + sin(r_y), rd.y, cos(r_y) - rd.x * sin(r_y));
            rd = normalize(rd);

            float3 f_i = (float3)(1.f, 1.f, 1.f);
            float3 nml = (float3)(1.f, 1.f, 1.f);
            float3 col = (float3)(0.f, 0.f, 0.f);
            float3 prp = (float3)(1.f, 0.f, 0.f);
            float3 mat = (float3)(0.f, 0.f, 0.f);
            float3 l_i = (float3)(1.f, 1.f, 1.f);
            for(int i = maxBounce - 1; i >= 0; i--){
                col = (float3)(0.f, 0.f, 0.f);
                float dist = 9999.f;
                float3 curve_pos = (float3)(22.f, c_y, 298.f);
                dist = intersect_quad_tex(  textures, texprops,   quad_pos,   quad_mat,   quad_prp,   quad_tex,   quads, &nml, &col, &prp, ro, rd, dist);
                dist = intersect_sphere_tex(textures, texprops, sphere_pos, sphere_mat, sphere_prp, sphere_tex, spheres, &nml, &col, &prp, ro, rd, dist);
                dist = intersect_collapse_tex(textures, texprops,  curve_pos, c_w, c_h, &nml, &col, &prp, ro, rd, dist);

                ro = ro + dist * rd;
                rd = reflect(rd, nml);
                ro = ro + 0.01 * rd;

                //l_i = light_rays(sphere_pos, quad_pos, light_pos, light_col, spheres, quads, lights, ro);
                l_i = light_rays(sphere_pos, quad_pos, light_pos, light_col, spheres, quads, lights, ro, c_w, c_h, c_y);
                mat = mat + prp.x * col * f_i * l_i;
                f_i = f_i * l_i * prp.y;
            }
            aa_cols += mat;
        }
    }
    unsigned char r = 31 * clamp(aa_cols.x / 4.0, 0.0, 1.0);
    unsigned char g = 63 * clamp(aa_cols.y / 4.0, 0.0, 1.0);
    unsigned char b = 31 * clamp(aa_cols.z / 4.0, 0.0, 1.0);
    pixels[idx] = ( r << 11 ) | ( g << 5 ) | b;
}
