#define WIDTH  320
#define HEIGHT 240
#define DX 4 / WIDTH / 3
#define DY 1 / HEIGHT

float3 reflect(float3 r, float3 n)
{
    return r - 2*(dot(r,n))*n;
}

float intersect_quad_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    __global __read_only float*  quad_pos,
    __global __read_only float*  quad_mat,
    __global __read_only float*  quad_prp,
    __global __read_only int* quad_tex,
    int quads,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < quads; i++){
        float3  x = (float3)(quad_pos[i * 14 + 0], quad_pos[i * 14 + 1], quad_pos[i * 14 + 2]);
        float3  n = (float3)(quad_pos[i * 14 + 3], quad_pos[i * 14 + 4], quad_pos[i * 14 + 5]);
        float3 e1 = (float3)(quad_pos[i * 14 + 6], quad_pos[i * 14 + 7], quad_pos[i * 14 + 8]);
        float3 e2 = (float3)(quad_pos[i * 14 + 9], quad_pos[i * 14 +10], quad_pos[i * 14 +11]);
        float w = quad_pos[i * 14 + 12];
        float h = quad_pos[i * 14 + 13];

        float tmp = dot(n, rd);
        float t = (tmp != 0.f) ? dot(n, (x - ro)) / tmp : 9000.f;
        if(t < dist && t > 0.f){
            float3 e = ro + t * rd - x;
            float p1 = dot(e, e1) / w;
            float p2 = dot(e, e2) / h;
            if((0.f <= p1 && p1 <= w) && (0.f <= p2 && p2 <= h)){
                dist = t;
                *nml = n;
                *prp = (float3)(quad_prp[i * 3], quad_prp[i * 3 + 1], quad_prp[i * 3 + 2]);
                if(quad_tex[3 * i] == -1)
                    *col = (float3)(quad_mat[i * 3], quad_mat[i * 3 + 1], quad_mat[i * 3 + 2]);
                else{

                    size_t offset = texprops[3 * quad_tex[3 * i]];
                    size_t u, v;
                    if(quad_prp[i * 3 + 2] < 0.5){
                        u = (size_t)(p1 * 4) % quad_tex[3 * i + 1];
                        v = (size_t)(p2 * 4) % quad_tex[3 * i + 2];
                    }
                    else{
                        u = (size_t)(clamp(p1 / w, 0.0, 1.0) * quad_tex[3 * i + 1]);
                        v = (size_t)(clamp(p2 / h, 0.0, 1.0) * quad_tex[3 * i + 2]);
                    }
                    offset += 3 * (v * quad_tex[3 * i + 1] + u);
                    *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
                    /*
                    size_t u, v;
                    size_t offset = texprops[3 * quad_tex[3 * i + 0]];
                    u = (size_t)(p1 * 4.f) % quad_tex[3 * i + 1];
                    v = (size_t)(p2 * 4.f) % quad_tex[3 * i + 2];
                    offset = offset + 3 * (v * quad_tex[3 * i + 1] + u);
                    */
                }
            }
        }
    }
    return dist;
}

float intersect_quad(
    __global __read_only float*  quad_pos,
    int quads,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < quads; i++){
        float3 x = (float3)(quad_pos[i * 14 + 0], quad_pos[i * 14 + 1], quad_pos[i * 14 + 2]);
        float3 n = (float3)(quad_pos[i * 14 + 3], quad_pos[i * 14 + 4], quad_pos[i * 14 + 5]);
        float3 e1 = (float3)(quad_pos[i * 14 + 6], quad_pos[i * 14 + 7], quad_pos[i * 14 + 8]);
        float3 e2 = (float3)(quad_pos[i * 14 + 9], quad_pos[i * 14 +10], quad_pos[i * 14 +11]);
        float w = quad_pos[i * 14 + 12];
        float h = quad_pos[i * 14 + 13];

        float tmp = dot(n, rd);
        float t = (tmp != 0.f) ? dot(n, (x - ro)) / tmp : 9000.f;
        if(t < dist && t > 0.f){
            float3 v = ro + t * rd - x;
            float p1 = dot(v, e1) / w;
            float p2 = dot(v, e2) / h;
            if((0.f <= p1 && p1 <= w) && (0.f <= p2 && p2 <= h)){
                dist = t;
            }
        }
    }
    return dist;
}

float intersect_sphere_tex(
    __global __read_only float*  textures,
    __global __read_only size_t* texprops,
    __global __read_only float* sphere_pos,
    __global __read_only float* sphere_mat,
    __global __read_only float* sphere_prp,
    __global __read_only int*   sphere_tex,
    int spheres,
    float3* nml,
    float3* col,
    float3* prp,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < 6; i++){
        float3 x = (float3)(sphere_pos[i * 4 + 0], sphere_pos[i * 4 + 1], sphere_pos[i * 4 + 2]);
        float r = sphere_pos[i * 4 + 3];
        float3 rc = ro - x;
        float c = dot(rc, rc) - (r*r);
        float b = dot(rd, rc);
        float d = b*b - c;
        float t = -b - sqrt(fabs(d));
        float s = step(0.0, min(t,d)) * step(t, dist);
        if(t < dist && min(t,d) > 0.f){
            float3 p = ro + rd * t - x;
            dist = t;
            *nml = -p / r;
            p /= r;
            //*nml = (x - (ro + rd * t)) / r;
            *prp = (float3)(sphere_prp[i * 3 + 0], sphere_prp[i * 3 + 1], sphere_prp[i * 3 + 2]);
            if(sphere_tex[3 * i] == -1)
                *col = (float3)(sphere_mat[i * 3 + 0], sphere_mat[i * 3 + 1], sphere_mat[i * 3 + 2]);
            else{
                float ru = (1.f - (atan2(p.z, p.x) + 3.14159265f) / 6.283185f);
                float rv = ((asin(p.y) + 1.570796326f) / 3.14159265f);
                ru = clamp(ru, 0.001, 0.999);
                rv = clamp(rv, 0.001, 0.999);
                size_t u = (size_t)(ru * sphere_tex[3 * i + 1]);
                size_t v = (size_t)(rv * sphere_tex[3 * i + 2]);
                size_t offset = texprops[3 * sphere_tex[3 * i]];
                offset += 3 * (v * sphere_tex[3 * i + 1] + u);
                *col = (float3)(textures[offset + 0], textures[offset + 1], textures[offset + 2]);
            }
        }
    }
    return dist;
}

float intersect_sphere(
    __global __read_only float* sphere_pos,
    int spheres,
    float3 ro,
    float3 rd,
    float dist
){
    for(int i = 0; i < 6; i++){
        float3 x = (float3)(sphere_pos[i * 4 + 0], sphere_pos[i * 4 + 1], sphere_pos[i * 4 + 2]);
        float r = sphere_pos[i * 4 + 3];
        float3 rc = ro - x;
        float c = dot(rc, rc) - (r*r);
        float b = dot(rd, rc);
        float d = b*b - c;
        float t = -b - sqrt(fabs(d));
        float s = step(0.0, min(t,d)) * step(t, dist);
        if(t < dist && min(t,d) > 0.f){
            dist = t;
        }
    }
    return dist;
}


float3 light_rays(
    __global __read_only float* sphere_pos,
    __global __read_only float* quad_pos,
    __global __read_only float* light_pos,
    __global __read_only float* light_col,
    int spheres, int quads, int lights,
    float3 ro
){
    float3 nil = (float3)(0, 0, 0);
    float3 l_i = (float3)(0, 0, 0);
    for(int i = 0; i < lights; i++){
        float dist = 9999.f;
        float3 lo = (float3)(light_pos[i * 3 + 0], light_pos[i * 3 + 1], light_pos[i * 3 + 2]);
        float3 lc = (float3)(light_col[i * 3 + 0], light_col[i * 3 + 1], light_col[i * 3 + 2]);
        float3 ld = normalize(lo - ro);
        dist = intersect_quad(quad_pos, quads, ro, ld, dist);
        dist = intersect_sphere(sphere_pos, spheres, ro, ld, dist);

        float3 d = ro - lo;
        float x = sqrt(dot(d, d));
        float s = step(x, dist);
        l_i = l_i + mix((float3)(0.0f, 0.00f, 0.00f), 1.25 * lc * pow(0.99, x), s);
    }
    l_i = clamp(l_i, 0.1, 1.0);
    return l_i;
}

// Super Sampling Anti-Aliasing Version (2x ssaa, no rotation or other special sampling technique)
__kernel void ray_tracer(
    __global __write_only short* pixels,
    __global __read_only float* textures,
    __global __read_only size_t* texprops,
    __global __read_only float* sphere_pos,
    __global __read_only float* sphere_mat,
    __global __read_only float* sphere_prp,
    __global __read_only int* sphere_tex,
    __global __read_only float* quad_pos,
    __global __read_only float* quad_mat,
    __global __read_only float* quad_prp,
    __global __read_only int* quad_tex,
    __global __read_only float* light_pos,
    __global __read_only float* light_col,
    int spheres, int quads, int lights,
    float x_x, float x_y, float x_z,
    float r_x, float r_y, float r_z,
    int maxBounce
)
{
    int x = get_global_id(0);
    int y = get_global_id(1);
    int idx = x + y * WIDTH;
    float3 aa_cols = (float3)(0.f, 0.f, 0.f);
    for(int ii = 0; ii < 2; ii++){
        for(int jj = 0; jj < 2; jj++){
            float3 ro = (float3)(x_x, x_y, x_z);
            float3 rd = (float3)(-1.f + .99f*DX*(2.0*x + ii), -1.f + .99f*DY*(2.0*y + jj), 1.f);

            rd = (float3)(rd.x * cos(r_y) + sin(r_y), rd.y, cos(r_y) - rd.x * sin(r_y));
            rd = normalize(rd);

            float3 f_i = (float3)(1.f, 1.f, 1.f);
            float3 nml, col, mat = (float3)(0.f, 0.f, 0.f);
            float3 prp, l_i;
            for(int i = maxBounce - 1; i >= 0; i--){
                col = (float3)(0.f, 0.f, 0.f);
                float dist = 9999.f;
                dist = intersect_quad_tex(  textures, texprops,   quad_pos,   quad_mat,   quad_prp,   quad_tex,   quads, &nml, &col, &prp, ro, rd, dist);
                dist = intersect_sphere_tex(textures, texprops, sphere_pos, sphere_mat, sphere_prp, sphere_tex, spheres, &nml, &col, &prp, ro, rd, dist);

                ro = ro + dist * rd;
                rd = reflect(rd, nml);
                ro = ro + 0.0001 * rd;

                l_i = light_rays(sphere_pos, quad_pos, light_pos, light_col, spheres, quads, lights, ro);
                mat = mat + prp.x * col * f_i * l_i;
                f_i = f_i * l_i * prp.y;
            }
            aa_cols += mat;
        }
    }
    unsigned char r = 31 * clamp(aa_cols.x / 4.0, 0.0, 1.0);
    unsigned char g = 63 * clamp(aa_cols.y / 4.0, 0.0, 1.0);
    unsigned char b = 31 * clamp(aa_cols.z / 4.0, 0.0, 1.0);
    pixels[idx] = ( r << 11 ) | ( g << 5 ) | b;
}
